@@ -2,17 +2,22 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
-from .work import *
-from .timesheet import *
+from . import ir
+from . import work
+from . import timesheet
 from . import party
 
 
 def register():
     Pool.register(
-        Work,
-        TimesheetWork,
+        # Before Work because status default value is read from WorkStatus
+        work.WorkStatus,
+        work.Work,
+        timesheet.Line,
+        timesheet.Work,
+        ir.ActWindow,
         module='project', type_='model')
     Pool.register(
-        party.PartyReplace,
-        party.PartyErase,
+        party.Replace,
+        party.Erase,
         module='project', type_='wizard')
