@@ -5,13 +5,117 @@ import datetime
 
 from sql import Null
 
-from trytond.model import ModelView, ModelSQL, fields, sequence_ordered, tree
-from trytond.pyson import Eval
+from trytond.cache import Cache
+from trytond.i18n import gettext
+from trytond.model import (
+    ModelView, ModelSQL, fields, sequence_ordered, tree, DeactivableMixin)
 from trytond.transaction import Transaction
 from trytond.pool import Pool
+from trytond.pyson import Eval, If, Bool, PYSONEncoder
 from trytond.tools import reduce_ids, grouped_slice
 
-__all__ = ['Work']
+from .exceptions import WorkProgressValidationError
+
+
+class WorkStatus(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
+    'Work Status'
+    __name__ = 'project.work.status'
+
+    _get_default_status_cache = Cache('project_work_status.get_default_status')
+    _get_window_domains_cache = Cache('project_work_status.get_window_domains')
+
+    types = fields.MultiSelection(
+        'get_types', "Types",
+        help="The type of works which can use this status.")
+    name = fields.Char("Name", required=True, translate=True)
+    progress = fields.Float(
+        "Progress",
+        domain=['OR',
+            ('progress', '=', None),
+            [
+                ('progress', '>=', 0),
+                ('progress', '<=', 1),
+                ],
+            ],
+        help="The minimum progress required for this status.")
+    default = fields.Boolean(
+        "Default",
+        help="Check to use as default status for the type.")
+    count = fields.Boolean(
+        "Count",
+        help="Check to show the number of works in this status.")
+
+    @classmethod
+    def get_types(cls):
+        pool = Pool()
+        Work = pool.get('project.work')
+        return Work.fields_get(['type'])['type']['selection']
+
+    @classmethod
+    def get_default_status(cls, type_=None):
+        if type_ is None:
+            return None
+        status = cls._get_default_status_cache.get(type_, -1)
+        if status != -1:
+            return status
+        records = cls.search([
+                ('types', 'in', type_),
+                ('default', '=', True)
+                ], limit=1)
+        if records:
+            status = records[0].id
+        else:
+            status = None
+        cls._get_default_status_cache.set(type, status)
+        return status
+
+    @classmethod
+    def create(cls, vlist):
+        cls._get_default_status_cache.clear()
+        cls._get_window_domains_cache.clear()
+        return super().create(vlist)
+
+    @classmethod
+    def write(cls, *args):
+        super().write(*args)
+        cls._get_default_status_cache.clear()
+        cls._get_window_domains_cache.clear()
+
+    @classmethod
+    def delete(cls, status):
+        cls._get_default_status_cache.clear()
+        cls._get_window_domains_cache.clear()
+        super().delete(status)
+
+    @classmethod
+    def get_window_domains(cls, action):
+        pool = Pool()
+        Data = pool.get('ir.model.data')
+        if action.id == Data.get_id('project', 'act_project_tree'):
+            return cls._get_window_domains([x[0] for x in cls.get_types()])
+        elif action.id == Data.get_id('project', 'act_project_form'):
+            return cls._get_window_domains(['project'])
+        elif action.id == Data.get_id('project', 'act_task_form'):
+            return cls._get_window_domains(['task'])
+        else:
+            return []
+
+    @classmethod
+    def _get_window_domains(cls, types):
+        key = tuple(sorted(types))
+        domains = cls._get_window_domains_cache.get(key)
+        if domains is not None:
+            return domains
+        encoder = PYSONEncoder()
+        domains = []
+        for status in cls.search([('types', 'in', types)]):
+            domain = encoder.encode([('status', '=', status.id)])
+            domains.append((status.name, domain, status.count))
+        if domains:
+            domains.append(
+                (gettext('project.msg_domain_all'), '[]', False))
+        cls._get_window_domains_cache.set(key, domains)
+        return domains
 
 
 class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
@@ -28,7 +132,11 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
     party = fields.Many2One('party.party', 'Party',
         states={
             'invisible': Eval('type') != 'project',
-            }, depends=['type'])
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['type', 'company'])
     party_address = fields.Many2One('party.address', 'Contact Address',
         domain=[('party', '=', Eval('party'))],
         states={
@@ -53,13 +161,13 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         'get_timesheet_date', setter='set_timesheet_date')
     timesheet_duration = fields.Function(fields.TimeDelta('Duration',
             'company_work_time',
-            help="Total time spent on this work and the sub-works"),
+            help="Total time spent on this work and the sub-works."),
         'get_total')
     effort_duration = fields.TimeDelta('Effort', 'company_work_time',
-        help="Estimated Effort for this work")
+        help="Estimated Effort for this work.")
     total_effort = fields.Function(fields.TimeDelta('Total Effort',
             'company_work_time',
-            help="Estimated total effort for this work and the sub-works"),
+            help="Estimated total effort for this work and the sub-works."),
         'get_total')
     progress = fields.Float('Progress',
         domain=['OR',
@@ -69,12 +177,13 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
                 ('progress', '<=', 1),
                 ],
             ],
-        help='Estimated progress for this work')
+        help='Estimated progress for this work.')
     total_progress = fields.Function(fields.Float('Total Progress',
             digits=(16, 4),
-            help='Estimated total progress for this work and the sub-works',
+            help='Estimated total progress for this work and the sub-works.',
             states={
-                'invisible': Eval('total_progress', None) == None,
+                'invisible': (
+                    Eval('total_progress', None) == None),  # noqa: E711
                 }),
         'get_total')
     comment = fields.Text('Comment')
@@ -91,10 +200,10 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
             ('company', '=', Eval('company', -1)),
             ],
         depends=['company'])
-    state = fields.Selection([
-            ('opened', 'Opened'),
-            ('done', 'Done'),
-            ], 'State', required=True, select=True)
+    status = fields.Many2One(
+        'project.work.status', "Status", required=True, select=True,
+        domain=[If(Bool(Eval('type')), ('types', 'in', Eval('type')), ())],
+        depends=['type'])
 
     @staticmethod
     def default_type():
@@ -104,9 +213,11 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
     def default_company(cls):
         return Transaction().context.get('company')
 
-    @staticmethod
-    def default_state():
-        return 'opened'
+    @classmethod
+    def default_status(cls):
+        pool = Pool()
+        WorkStatus = pool.get('project.work.status')
+        return WorkStatus.get_default_status(cls.default_type())
 
     @classmethod
     def default_left(cls):
@@ -119,7 +230,9 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TimesheetWork = Pool().get('timesheet.work')
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
         table_project_work = cls.__table_handler__(module_name)
         project = cls.__table__()
         timesheet = TimesheetWork.__table__()
@@ -138,9 +251,9 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         if table_project_work.column_exist('effort'):
             cursor.execute(*project.select(project.id, project.effort,
                     where=project.effort != Null))
-            for id_, effort in cursor.fetchall():
+            for id_, effort in cursor:
                 duration = datetime.timedelta(hours=effort)
-                cursor.execute(*project.update(
+                update.execute(*project.update(
                         [project.effort_duration],
                         [duration],
                         where=project.id == id_))
@@ -156,8 +269,8 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
                     condition=timesheet.parent == second_project.work
                     ).select(project.id, second_project.id)
             cursor.execute(*query)
-            for id_, parent in cursor.fetchall():
-                cursor.execute(*project.update(
+            for id_, parent in cursor:
+                update.execute(*project.update(
                         [project.parent],
                         [parent],
                         where=project.id == id_))
@@ -166,8 +279,8 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
             cursor.execute(*project.join(timesheet,
                     condition=project.work == timesheet.id
                     ).select(project.id, timesheet.company))
-            for id_, company in cursor.fetchall():
-                cursor.execute(*project.update(
+            for id_, company in cursor:
+                update.execute(*project.update(
                         [project.company],
                         [company],
                         where=project.id == id_))
@@ -176,8 +289,8 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
             cursor.execute(*project.join(timesheet,
                     condition=project.work == timesheet.id
                     ).select(project.id, timesheet.name))
-            for id_, name in cursor.fetchall():
-                cursor.execute(*project.update(
+            for id_, name in cursor:
+                update.execute(*project.update(
                         [project.name],
                         [name],
                         where=project.id == id_))
@@ -185,7 +298,6 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         # Migration from 4.0: remove work
         if work_exist:
             table_project_work.drop_constraint('work_uniq')
-            update = Transaction().connection.cursor()
             cursor.execute(*project.select(project.id, project.work,
                     where=project.work != Null))
             for project_id, work_id in cursor:
@@ -195,16 +307,24 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
                         where=timesheet.id == work_id))
             table_project_work.drop_column('work')
 
-    @classmethod
-    def __setup__(cls):
-        super(Work, cls).__setup__()
-        cls._error_messages.update({
-                'invalid_parent_state': ('Work "%(child)s" can not be opened '
-                    'because its parent work "%(parent)s" is already done.'),
-                'invalid_children_state': ('Work "%(parent)s" can not be '
-                    'done because its child work "%(child)s" is still '
-                    'opened.'),
-                })
+        # Migration from 5.4: replace state by status
+        table_project_work.not_null_action('state', action='remove')
+
+    @fields.depends('type', 'status')
+    def on_change_type(self):
+        pool = Pool()
+        WorkState = pool.get('project.work.status')
+        if (self.type
+                and (not self.status
+                    or self.type not in self.status.types)):
+            self.status = WorkState.get_default_status(self.type)
+
+    @fields.depends('status', 'progress')
+    def on_change_status(self):
+        if (self.status
+                and self.status.progress is not None
+                and self.status.progress > (self.progress or -1.0)):
+            self.progress = self.status.progress
 
     @classmethod
     def index_set_field(cls, name):
@@ -217,22 +337,33 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
     def validate(cls, works):
         super(Work, cls).validate(works)
         for work in works:
-            work.check_state()
+            work.check_work_progress()
 
-    def check_state(self):
-        if (self.state == 'opened'
-                and (self.parent and self.parent.state == 'done')):
-            self.raise_user_error('invalid_parent_state', {
-                    'child': self.rec_name,
-                    'parent': self.parent.rec_name,
-                    })
-        if self.state == 'done':
-            for child in self.children:
-                if child.state == 'opened':
-                    self.raise_user_error('invalid_children_state', {
-                            'parent': self.rec_name,
-                            'child': child.rec_name,
-                            })
+    def check_work_progress(self):
+        pool = Pool()
+        progress = -1 if self.progress is None else self.progress
+        if (self.status.progress is not None
+                and progress < self.status.progress):
+            Lang = pool.get('ir.lang')
+            lang = Lang.get()
+            raise WorkProgressValidationError(
+                gettext('project.msg_work_invalid_progress_status',
+                    work=self.rec_name,
+                    progress=lang.format('%.2f%%', self.status.progress * 100),
+                    status=self.status.rec_name))
+        if (self.status.progress == 1.0
+                and not all(c.progress == 1.0 for c in self.children)):
+            raise WorkProgressValidationError(
+                gettext('project.msg_work_children_progress',
+                    work=self.rec_name,
+                    status=self.status.rec_name))
+        if (self.parent
+                and self.parent.progress == 1.0
+                and not self.progress == 1.0):
+            raise WorkProgressValidationError(
+                gettext('project.msg_work_parent_progress',
+                    work=self.rec_name,
+                    parent=self.parent.rec_name))
 
     @property
     def effort_hours(self):
@@ -333,7 +464,7 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
             where = reduce_ids(table.id, sub_ids)
             cursor.execute(*table.select(table.id, table.parent,
                     where=where))
-            parents.update(cursor.fetchall())
+            parents.update(cursor)
 
         if 'total_progress' in names and 'total_effort' not in names:
             names = list(names)
@@ -350,8 +481,8 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
             total_effort = result['total_effort']
             for work in works:
                 if total_effort[work.id]:
-                    total_progress[work.id] = round(total_progress[work.id] /
-                        (total_effort[work.id].total_seconds() / 60 / 60),
+                    total_progress[work.id] = round(total_progress[work.id]
+                        / (total_effort[work.id].total_seconds() / 60 / 60),
                         digits)
                 else:
                     total_effort[work.id] = None
@@ -378,11 +509,16 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
 
     @classmethod
     def copy(cls, project_works, default=None):
+        pool = Pool()
+        WorkStatus = pool.get('project.work.status')
         if default is None:
             default = {}
         else:
             default = default.copy()
         default.setdefault('children', None)
+        default.setdefault('progress', None)
+        default.setdefault(
+            'status', lambda data: WorkStatus.get_default_status(data['type']))
         return super().copy(project_works, default=default)
 
     @classmethod
